/* -------------------------- Problem Details -------------------------- 
    This is a demo for solving the problem:
    
                min 0.5*x'*T*x + e'*x

    where T is the tridiagonal matrix with all diagonal elements 2 except the
    first element which is 1; the super and subdiagonal elements of T are all 1.    e is vector with all elements 1.

    The primary purpose of this demo is to show the user how to utilize
    cg_descent's specialized routines for handling a quadratic objective. */

#include "cg_descent.h"

/* prototypes */
void hprod
(
    CGFLOAT   *val,
    CGFLOAT     *x,
    CGINT        n
) ;

int main (void)
{
    int status = 0 ;
    CGINT i, j, n ;

    CGdata *cgdata = cg_setup () ; /* initialize a CGdata structure */

    n = cgdata->n = 10000 ; /* set problem dimension */
    /* since cgdata->x is not specified, the starting guess is x = 0 */

    /* For a quadratic objective, the linear cost vector is stored cgdata->c,
       an array of length n. We first need to malloc space for the elements of
       cgdata->c */
    cgdata->c = cg_malloc (&status, n, sizeof (CGFLOAT)) ;

    /* If status = SOPT_OUT_OF_MEMORY, then there was not enough memory
       for the malloc. For our problem, the elements of the linear
       cost vector were all one. There is a builtin function cg_initx
       for setting all the elements of an CGFLOAT array to fixed scalar. */
    cg_initx (cgdata->c, 1, n) ;
    
    /* For a quadratic objective, the Hessian is stored in cgdata->H, a
       pointer to an SOPT_matrix structure defined in SuiteOPTconfig/sopt.h. */
    SOPT_matrix *H = cgdata->H ;

    /* There are three possible formats that could be used: dense matrix,
       sparse matrix, or triples. If dense or sparse formats are used,
       then all the elements of the matrix must be given, and C-indexing
       must be used where the initial row and column are both 0. With the
       triples format, there are two options: If H->sym = TRUE (default),
       then the matrix is symmetric and only elements on the main diagonal
       and one side are given. If H->fortran = TRUE (default), then Fortran
       indexing is used where the initial row and column of a matrix are 1.
       We use triples with the default options to input the Hessian, which
       is the matrix T for our problem. Hence, the number of nonzeros
       in the Hessian is nnz = 2*n - 1 (n elements on the main diagaonal and
       n-1 elements on the superdiagonal. To begin, we malloc space for
       row and column indices and the nonzero matrix elements: */

    CGINT nnz = 2*n - 1 ; /* number of nonzeros in Hessian input */
    H->nnz  = nnz ;
    H->rows = (CGINT *)   cg_malloc (&status, nnz, sizeof (CGINT)) ;
    H->cols = (CGINT *)   cg_malloc (&status, nnz, sizeof (CGINT)) ;
    H->vals = (CGFLOAT *) cg_malloc (&status, nnz, sizeof(CGFLOAT));

    /* now input the matrix elements for the Hessian */
    CGINT n1 = n - 1 ;
    j = 0 ;
    /* input diagonal and adjacent superdiagaonal */
    for (i = 1; i <= n1; i++)
    {
        /* diagonal elements */
        H->rows [j] = i ; H->cols [j] = i ;     H->vals [j] =  2 ;
        j++ ;
        /* adjacent superdiagonal element */
        H->rows [j] = i ; H->cols [j] = i + 1 ; H->vals [j] =  1 ;
        j++ ;
    }
    /* correct special first diagonal element, add final last diagonal element*/
                                        H->vals [0] = 1 ;
    H->rows [j] = n ; H->cols [j] = n ; H->vals [j] = 2 ;

    /* by default, the run statistics are not printed. Set PrintStat to TRUE */
    cgdata->Parm->PrintStat = TRUE ;

    /* Run the code and time it. The initial call to the timer includes a
       startup cost associated with its initialization. Thus we call the
       timer twice initially, the first call absorbs the initialization time,
       while the second call starts the real timer. */
    CGFLOAT walltime = cg_timer () ;

    walltime = cg_timer () ; /* start the real timer */
    cg_descent (cgdata) ;
    walltime = cg_timer () - walltime ;

    printf ("Time of Run 1: %e\n", walltime) ;
    printf ("x [1:4] =\n") ;
    for (i = 0; i < 4; i++) printf ("    %15.5f\n", cgdata->x [i]) ;

    /* By default, the code will exploit the symmetric solver, when it is
       available. For Run 2, we apply the purely gradient-based method
       choosing deriv_mode = 1. */
    cgdata->Parm->deriv_mode = 1 ;

    /* To make the starting guess for Run 2 the same as the starting guess 0
       used in Run 1, we need set cgdata->x to 0 since this data element
       currently holds the previously computed optimal solution. */
    cg_initx (cgdata->x, 0, n) ;

    walltime = cg_timer () ; /* start the real timer */
    cg_descent (cgdata) ;
    walltime = cg_timer () - walltime ;

    printf ("Time of Run 2: %e\n", walltime) ;
    printf ("x [1:4] =\n") ;
    for (i = 0; i < 4; i++) printf ("    %15.5f\n", cgdata->x [i]) ;

    /* access the number of function and gradient evaluations from the
       Stat structure found in cgdata */
    printf ("Number of Function Evaluations: %li\n",
            (LONG) cgdata->Stat->nfunc) ;
    printf ("Number of Gradient Evaluations: %li\n",
            (LONG) cgdata->Stat->ngrad) ;

    /* free the memory allocated by the user for c */
    free (cgdata->c) ;

    /* free the workspace created by cg_setup, including the memory
       allocated by cg_descent for cgdata->x */
    cg_terminate (&cgdata) ;
}
