function demoHS
    %  Test problem: Hock and Schittkowski problem 112 as implemented in CUTEST
    %                problem HS112.SIF
    %
    %    min f(x) s.t. 1.e-6 <= x, A*x = [2 1 1 0]'
    %
    %  where
    %
    %    f(x) = sum_{i = 1 to 10} x_i(c_i + log (x_i) - log (x_{11}))
    %
    %         | 1  2  2  0  0  1  0  0  0  1  0|
    %     A = | 0  0  0  1  2  1  1  0  0  0  0|
    %         | 0  0  1  0  0  0  1  1  2  1  0|
    %         | 1  1  1  1  1  1  1  1  1  1 -1|
    %
    %     c = [-6.089 -17.164 -34.054  -5.914 -24.721 ...
    %         -14.986 -24.100 -10.708 -26.662 -22.179 ]
    %
    %  Starting guess: x_i = 0.1, 1 <= i <= 10, x_{11} = 1
    %
    %  In CUTEST problem HS112.SIF, the right side [2 1 0 0] of problem 112 is
    %  replaced by [2 1 1 0]. With this replacement, the solution given by
    %  Hock and Schittkowski for problem 112 is obtained.

    %  Currently, there are 2 different ways to give the elements of the
    %  constraint matrix A:

    %  (1) MATLAB sparse matrix
    %  (2) A dense matrix

    %  You could construct A using triples, but then convert A to sparse format
    %  using the sparse command. For example, the triple form of
    %  A = [ 0.1 0 0 ; 0.2 0 0.4] is T = [1 1 0.1 ; 2 1 0.2 ; 2 3 0.4] 
    %  and the sparse form of A if A = sparse (T) ;

    pasadata.A = [ 1, 2, 2, 0, 0, 1, 0, 0, 0, 1, 0,
                   0, 0, 0, 1, 2, 1, 1, 0, 0, 0, 0,
                   0, 0, 1, 0, 0, 0, 1, 1, 2, 1, 0,
                   1, 1, 1, 1, 1, 1, 1, 1, 1, 1,-1 ] ;

    % Input the lower bounds for A*x
    pasadata.bl = [2 1 1 0] ;
    % pasadata->bu = bl since the upper and lower bounds are equal
    pasadata.bu = pasadata.bl ;
    c = [-6.089 -17.164 -34.054  -5.914 -24.721 ...
        -14.986 -24.100 -10.708 -26.662 -22.179 ]' ;

    % the starting guess is the vector with all its entries 0.1 except for
    % x(11) = 1
    pasadata.x = 0.1*ones(11, 1) ;
    pasadata.x(11) = 1 ;

    % lower bounds is the vector with all its entries 1.e-6
    pasadata.lo = 1.e-6*ones(11, 1) ;

    % nothing is stored in pasadata->hi since there are no upper bounds

    % The routines to evaluate the function and the gradient appear below.
    pasadata.value = @value ;
    pasadata.grad = @grad ;

    % if a routine to evaluate the function value and gradient together were
    % provided, then it would be stored in pasadata.valgrad. The Hessian
    % also appears below. To utilize the Hessian-based implementation of pasa:

    pasadata.hessian = @hess ;

    % Even though the Hessian is provided, it may not be used by pasa
    % if the problem can be solved quicker using the purely gradient-based
    % approach. In particular, the Hessian is not used for this problem.
    % To force pasa to use the Hessian, set pasadata.pasa.deriv_mode = 2 ;
    % one finds that the solution time for the Hessian-based approach is
    % greater than the solution time for the gradient-based approach.

    pasadata.pasa.PrintStat = 1 ;

    tic
    x = pasa (pasadata) ; 
    toc

    disp ('x =') ;
    disp (x) ;

    function f = value(x)
    
        % Since the pasa iterates are feasible, x [10] = x [0] + ... + x [9]
        % Thus the x_{11} term in f reduces to -x(10)*log(x(10)
        f = x(1:10)'*(c + log(x(1:10))) - x(11)*log(x(11)) ;
    end
    
    function g = grad(x)
        t = 1 - log(x(11)) ;
        g(1:10) = c + log(x(1:10)) + t ;
        g(11) = -1 ;
    end
    
    % The nonzeros of the Hessian are on the main diagonal (1/x_i) and the
    % last row and column (-1/x(11). The element on the main diagonal and above
    % should be input using triples. 
    function H = hess(x)
        t = -1/x(11) ;
        H = [ [1:11]', [1:11]', 1./x
              [1:10]', 11*ones(10, 1), t*ones(10, 1)] ;
    end
end
