/* -------------------------- Problem Details -------------------------- 
    This is a demo for solving the problem
 
                 min 0.5*(n^2)*x'*T*x + sum_{i=1}^{n-1} exp (-x_i)
 
    where T is the tridiagonal matrix with 2's on the diagonal and -1's on the
    super and subdiagonals. The gradient g of the objective is
    (n^2)*T*x - exp(-x), where exp(-x) is the column vector with i-th
    element exp(-x_i). At the minimizer of the objective, the
    gradient vanishes and (n^2)*T*x =  exp(-x). This equation is the
    central difference discretization of the boundary-value problem:
    -x''(t) = exp(-x(t)), x(0) = x(1) = 0. Consequently, if x is the solution
    to the optimization problem, then x_i is an approximation to the solution
    to the boundary-value problem at t = i/n. */

#include "cg_descent.h"

/* prototypes */
void value
(
    CGFLOAT *f,
    CGFLOAT *x,
    CGINT    n
) ;

void grad
(
    CGFLOAT *g,
    CGFLOAT *x,
    CGINT    n
) ;

void hess
(
    SOPT_matrix *H,
    SOPTFLOAT   *x,
    SOPTINT      n
) ;

/* global variables */
CGINT n2, *rows, *cols ;
CGFLOAT nn, *vals ;

int main (void)
{
    CGINT i, n ;

    /* A CGdata structure is used to input the problem formulation.
       The fields of the structure are initialized as NULL or EMPTY */
    CGdata *cgdata = cg_setup () ;

    /* Choose a value for the parameter in the problem formulation */
    n = 2000 ;
    /* Other problem parameters (global) */
    nn = (double) n * (double) n ;
    n2 = n - 2 ;

    /* Insert dimension of x in the input data structure */
    cgdata->n = n - 1 ;

    /* If an initial guess for the problem solution is not provided in
       cgdata->x, then cg_descent will malloc the space for the starting
       guess in cgdata->x and will set the starting guess to x = zero. */

    /* The stopping condition for pasa is ||g|| <= cgdata->Parm->grad_tol,
       where ||g|| denotes the sup-norm of the gradient at the current iterate.
       The default value for grad_tol is 1.e-6.  */

    /* Input the name of the routines for evaluating the objective and its
       gradient in the cgdata structure: */
    cgdata->value = value ;
    cgdata->grad  = grad ;

    /* Print statistics for the run by setting PrintStat = TRUE; by default,
       PrintStat = FALSE. As an alternative to setting the PrintStat parameter
       to TRUE, the user could print all the statistics after the run is
       complete using the statement "cg_print_stat (cgdata) ;" */
    cgdata->Parm->PrintStat = TRUE ;

    /* Run the code and time it. The initial call to the timer includes a
       startup cost associated with its initialization. Thus we call the
       timer twice initially, the first call absorbs the initialization time,
       while the second call starts the real timer. */
    CGFLOAT walltime = cg_timer () ;

    walltime = cg_timer () ; /* start the real timer */
    cg_descent (cgdata) ;
    walltime = cg_timer () - walltime ;

    printf ("Time of Run 1: %e\n", walltime) ;
    printf ("Print first 4 components of x:\n") ;
    for (i = 0; i < 4; i++) printf ("  %10.7f\n", cgdata->x [i]) ;
    printf ("\n") ;

    /* We now resolve the problem from start guess x = 0 using the Hessian-based
       version of pasa. Since cgdata->x currently holds the computed solution,
       we initialize cgdata->x = 0 in the following statement. */
    cg_initx (cgdata->x, CGZERO, n-1) ;

    /* By inserting the name of the Hessian evaluation routine in the
       cgdata structure, pasa will by default exploit the Hessian when solving
       the problem. */
    cgdata->hessian = hess ;

    walltime = cg_timer () ;
    cg_descent (cgdata) ;
    walltime = cg_timer () - walltime ;

    printf ("Time of Run 2 (using Hessian): %e\n", walltime) ;
    printf ("Print first 4 components of x:\n") ;
    for (i = 0; i < 4; i++) printf ("  %10.7f\n", cgdata->x [i]) ;
    printf ("\n") ;

    /* Below we print the number of function and gradient evaluations found
       in the statistics structure. */
    printf ("Number of Function Evaluations from the Stat structure: %li\n",
            (LONG) cgdata->Stat->nfunc) ;
    printf ("Number of Gradient Evaluations from the Stat structure: %li\n",
            (LONG) cgdata->Stat->ngrad) ;

    /* Finally, we free the workspace created by cg_setup, including the memory
       allocated by cg_descent for cgdata->x */
    cg_terminate (&cgdata) ;

    /* free the memory created for the triples arrays in Hessian routine hess */
    free (rows) ;
    free (cols) ;
    free (vals) ;
}

void value
(
    CGFLOAT *f,
    CGFLOAT *x,
    CGINT    n
)
{
    CGFLOAT t, s ;
    CGINT i ;
    /* In the problem formulation, the variables are x_1, x_2, ... while in C,
       the variable are x [0], x [1], ... */
    t = 2*x [0]*x [0] ;
    s = exp (-x [0]) ;
    for (i = 1; i <= n2; i++)
    {
        t += 2*(x [i]*x [i] - x [i]*x [i-1]) ;
        s += exp(-x [i]) ;
    }
    *f = 0.5*nn*t + s ;
}

/* ---- Gradient of objective function g = nn*T*x - exp(-x) --- */
void grad
(
    CGFLOAT *g,
    CGFLOAT *x,
    CGINT    n
)
{
    CGFLOAT t ;
    CGINT i ;
    g [0]  = nn*(2*x [0] - x [1]) - exp (-x [0]) ;
    g [n2] = nn*(2*x [n2] - x [n2-1]) - exp (-x [n2]) ;
    for (i = 1; i < n2; i++)
    {
        g [i] = nn*(2*x [i] - x [i-1] - x [i+1]) - exp (-x [i]) ;
    }
}


/* There are 3 ways to provide the Hessian; sparse matrix format,
   dense matrix, and triples. Here we use triples format where the
   potentially nonzero elements of the Hessian are specified using the
   row and column indices and the Hessian values for element on the
   main diagonal and above. For our problem, the Hessian is
   nn*T + diag(exp (-x)). */
void hess
(
    SOPT_matrix *H,
    SOPTFLOAT   *x,
    SOPTINT      n
)
{
    CGINT i, j ;
    /* At the first call to hess, malloc arrays and save the row and column
       indices of the nonzeros. This is only done one time. In subsequent call
       to hess, we only provide the elements of the Hessian that could change
       when x changes.  The number of nonzeros elements to be provided
       is 2*n - 3 = 2*n2 + 1 corresponding to the main diagonal (n-1) and the
       superdiagonal (n-2). */
    if ( H->nnz == EMPTY )
    {
        CGINT nnz ;
        nnz = 2*n2 + 1 ;
        H->rows = rows = malloc (nnz*sizeof (CGINT)) ;
        H->cols = cols = malloc (nnz*sizeof (CGINT)) ;
        H->vals = vals = malloc (nnz*sizeof (CGFLOAT)) ;
        j = 0 ;
        for (i = 0; i < n2; i++)
        {
            /* store the diagonal element */
            rows [j] = i ;
            cols [j] = i ;
            vals [j] = 2*nn + exp (-x [i]) ;
            j++ ;
            /* store the super diagonal element */
            rows [j] = i ;
            cols [j] = i + 1 ;
            vals [j] = -1*nn ; 
            j++ ;
        }
        /* store the final element on the diagonal */
        rows [j] = n2 ;
        cols [j] = n2 ;
        vals [j] = 2*nn + exp (-x [j]) ;
        H->sym = TRUE ; /* the Hessian is symmetric */
        H->nnz = nnz ;  /* the number of nonzeros on diagonal and above */
        H->fortran = FALSE ;
        /* NOTE: Above, we use C-indexing for the Hessian matrix where the first
                 row and column are 0. If Fortran indexing is used where the 1st
                 row and column are 1, then set H->fortran = TRUE. */
    }
    else /* update the elements of the Hessian that change when x changes */
    {
        vals = H->vals ;
        j = 0 ;
        /* the only elements that change are the diagonal elements */
        for (i = 0; i <= n2; i++) /* i = n2 is handled in the for loop */
        {
            vals [j] = 2*nn + exp (-x [i]) ;
            j += 2 ;
        }
    }
}
