function demo 
%  -------------------------- Problem Details -------------------------- 
%   This is a demo for solving the problem
 
%                min 0.5*(n^2)*x'*T*x + sum_{i=1}^{n-1} exp (-x_i)
 
%   where T is the tridiagonal matrix with 2's on the diagonal and -1's on the
%   super and subdiagonals. The gradient g of the objective is
%   (n^2)*T*x - exp(-x), where exp(-x) is the column vector with i-th
%   element exp(-x_i). At the minimizer of the objective, the
%   gradient vanishes and (n^2)*T*x =  exp(-x). This equation is the
%   central difference discretization of the boundary-value problem:
%   -x''(t) = exp(-x(t)), x(0) = x(1) = 0. Consequently, if x is the solution
%   to the optimization problem, then x_i is an approximation to the solution
%   to the boundary-value problem at t = i/n.

%% ----------- User defined parameter values for cg_descent ------------ %%
%   See cg_default inside CGDESCENT/Source/cg_default.c for a 
%   description of the parameters.
    cgdata.PrintStat = 1 ;    % print cg_descent statistics (default: no print)
    cgdata.grad_tol = 1.e-8 ; % 1e-6 is default if stopping tolerance not given

%% ------------- Calling cg_descent (Non-Quadratic Cost) --------------- %%
    % ----------------------- Required Inputs ------------------------ %
    n = 2000 ; nn = n*n ;    % Parameters in problem formulation
    n2 = n - 2 ; n1 = n - 1 ;% Other handy constants
    cgdata.n = n1 ;          % Store dimension of x in the problem data
    cgdata.value = @value ;  % Store function handle for objective
    cgdata.grad  = @grad ;   % Store function handle for gradient of obj
    % Since the starting guess cgdata.x is not specified, the default
    % starting guess is x = 0. If the starting guess is provided,
    % then pasa assumes that the length of cgdata.x is the dimension of x.

    % ----------------------- Call cg_descent ----------------------- %
       % The only required output is the solution. Optionally, the run
       % status and the cg_descent statistics are returned as an output. Above
       % we set cgdata.PrintStat = 1 (TRUE) so these statistics are printed.
    tic
    x = cg_descent (cgdata) ;
    toc

%% --------------------- Plot the solution x --------------------------- %%
    plot ([1:n1]/n, x) ;

    % ----------- Provide the Hessian and resolve the problem --------
    display ('Now provide the Hessian.') ;

    cgdata.hessian = @hess ;

    % since the Hessian has been provided, the code will exploit it to
    % accelerate convergence
    tic ;
    x = cg_descent (cgdata) ;
    toc

%% ----- User defined functions for cg_descent (Non-Quadratic Cost) ---- %%
    % ---- Objective function ---- %
    function f = value(x)
       f = exp(-x(1)) ;
       t = 2*x(1)^2 ;
       for i = 2:n1
           t = t + 2*(x(i)^2 - x(i)*x(i-1)) ;
           f = f + exp(-x(i)) ;
       end
       f = 0.5*nn*t + f ;
    end
    
    % ---- Gradient of objective function g = nn*T*x - exp(-x) ---- %
    function g = grad(x)
        % initialize ouput g
        g = zeros (n1, 1) ;
        g(1)  = nn*(2*x(1)  - x(2)) - exp(-x(1)) ;
        g(n1) = nn*(2*x(n1) - x(n2)) - exp(-x(n1)) ;
        for i = 2:n2
           g(i) = nn*(2*x(i) - x(i-1) - x(i+1)) - exp(-x(i)) ;
        end
    end

    % ---- Hessian of objective = nn*T + diag (exp(-x)) ---- %
    function H = hess(x)
        % Hessian is symmetric, only give the elements on the diagonal and
        % above the diagaonal that could be nonzero using a triples matrix.
        % Each row of the matrix give the row index, column index, and the
        % numerical value of a matrix element

        hdiag = 2*nn + exp(-x) ;
        H = [[1:n1]', [1:n1]',      hdiag         % diagonal entries
             [1:n2]', [2:n1]', -nn*ones(n2, 1)] ; % super diagonal entries
        % The first n - 1 rows of H correspond to the diagonal of H and
        %      last n - 2 rows of H correspond to the super diagonal
        % H = [ 1   1   2*nn + exp(-x(1))
        %       2   2   2*nn + exp(-x(2))
        %       .   .         .
        %      n1  n1   2*nn + exp(-x(n1))
        %       1   2       -1*nn
        %       2   3       -1*nn
        %       .   .        .
        %       n2 n1       -1*nn ]
    end
end
