function demoQP 
%% -------------------------- Problem Details -------------------------- %%
%   This is a demo for solving the problem
%   
%                min     sum_{i=1}^n  0.5*i*x_i^2 + x_i
%   
%   with the purpose of illustrating how to set up and solve a problem
%   using the quadratic mode of cg_descent. 

%   There are three different ways to input the Hessian matrix H:
%       1. As a function p = hprod(x) which accepts a vector x and which
%          outputs the product p = H*x; note that this approach for inputting
%          a quadratic causes cg_descent to employ its gradient-based
%          implementation since the Hessian-based implementation needs the
%          elements of the Hessian.
%       2. As a sparse matrix in triples format
%       3. As a dense matrix (the dense matrix is converted to sparse matrix
%          inside pasa)
%
%   These approaches are illustrated in this order.

%% ------------------ Initialize inputs for cg_descent ----------------- %%
    n = 10000 ;      % dimension of primal variable
    x = ones(1, n) ; % initial guess
    c = ones(1, n) ; % linear term in objective function
    H = [1:n]' ;
    
%% ------------- User defined parameter values for cg_descent ---------- %%
%   See cg_default inside CGDESCENT/Source/cg_default.c for a 
%   description of the parameters.
    cgdata.PrintStat = 1 ;   % print CGDESCENT statistics

%% ---------------- Calling cg_descent (Quadratic Cost) ---------------- %%
    % ----------------- Set options for cg_descent ------------------ %
    cgdata.x = x ; % Initial guess for solution
    cgdata.c = c ; % Linear term in cost function 
    cgdata.hprod = @myhprod ; % Function handle for hessian times vector 

    % ----------------------- Call cg_descent ----------------------- %
    tic
    X = cg_descent (cgdata) ;
    toc
%% --------------------- Print cg_descent outputs ---------------------- %%
    % Since cgdata.PrintStat = 1, the cg_descent statistics are printed
    % at the conclusion of the run. The numerical values for the statistics
    % are stored in the stats structure.

    % print solution x
    fprintf('\n  -------------------- Solution Using hprod')
    fprintf('-------------------\n')

    fprintf('X (1:4) =') 
    disp (X(1:4)) ;

%% --------------------- Input H as a sparse matrix --------------------- %%
    % Previously we input the Hessian/vector product as an element of cgdata.
    % Now remove it:
    cgdata = rmfield(cgdata,'hprod') ;  % remove hprod from cgdata
    % Hessian now input as a sparse diagonal matrix with diagonal H = [1:n]'
    cgdata.H = spdiags (H, [0], n, n) ;

    % ----------------------- Call cg_descent ----------------------- %
    tic
    X = cg_descent (cgdata) ;
    toc
    disp ('This is faster than using the product input since cg_descent') ;
    disp ('can exploit the sparse symmetric equation solver.') ;

    % print solution x
    fprintf('\n  ---------------- Solution Using Sparse Hessian')
    fprintf('---------------\n')

    fprintf('X (1:4) =') 
    disp (X(1:4)) ;

%% --------------------- Input H as a dense matrix --------------------- %%
    cgdata.H = diag (H, [0]) ; % Hessian is a dense matrix with H = [1:n]'
                               % on the diagonal and off-diagonal elements 0

    % ----------------------- Call cg_descent ----------------------- %
    tic
    X = cg_descent (cgdata) ;
    toc
    disp ('Most of time is spent converting dense matrix to sparse matrix');

    % print solution x
    fprintf('\n  ---------------- Solution Using Dense Hessian')
    fprintf('---------------\n')

    fprintf('X (1:4) =') 
    disp (X(1:4)) ;

%% -------- User defined functions for cg_descent (Quadratic Cost) ----- %%
    % ---- Product of Hessian Times vector ---- %
    function p = myhprod (x)
        % multiply each Hessian diagonal element by corresponding element of x
        p = H.*x ; % H = [1:n]'
    end
end
